using System;
using System.ComponentModel;
using System.Collections;
using System.Text;
using System.Data;
using System.Windows.Forms;
using Syncfusion.Windows.Forms.Grid;

namespace WrapperClass
{
    public class ListWrapper : ArrayList, ITypedList
    {
        internal PropertyDescriptorCollection pdc;
        IList originalList = null;
      
        #region Constructors
        public ListWrapper(IList list)
        {
            if(list.Count == 0)
                throw new Exception("Empty lists can only be wrapped using the ListWrapper(IList list, Type t) constructor.");
            pdc = TypeDescriptor.GetProperties(list[0].GetType());
            
            SetList(list);
            this.originalList = list;
        }
        public ListWrapper(IList list, Type t)
        {
            pdc = TypeDescriptor.GetProperties(t);
            SetList(list);
            this.originalList = list;
        }
        public ListWrapper(DataTable dataTable)
        {
            ProcessDataTable(new DataView(dataTable));
        }

        public ListWrapper(DataView dataView)
        {
            ProcessDataTable(dataView);
        }

        private void ProcessDataTable(DataView dataView)
        {
           pdc = ((ITypedList)dataView).GetItemProperties(null);
            SetList(dataView);
            this.originalList = dataView;
        }
        #endregion

		#region grid interaction

		GridSelectionFlags saveFlags;
		GridSortBehavior saveBehavior;
		GridDataBoundGrid grid = null;
		public void WireGrid(GridDataBoundGrid grid)
		{
			if(this.grid != null)
				UnWireGrid(this.grid);
			this.grid = grid;
			grid.CellClick += new Syncfusion.Windows.Forms.Grid.GridCellClickEventHandler(gridDataBoundGrid1_CellClick);
			grid.Model.QueryCellInfo += new GridQueryCellInfoEventHandler(Model_QueryCellInfo);
			saveFlags = grid.AllowSelection;
			saveBehavior = grid.SortBehavior;
			grid.AllowSelection &= ~GridSelectionFlags.Column;
			grid.SortBehavior = GridSortBehavior.SingleClick;
		}

		public void UnWireGrid(GridDataBoundGrid grid)
		{
			grid.CellClick -= new Syncfusion.Windows.Forms.Grid.GridCellClickEventHandler(gridDataBoundGrid1_CellClick);
			grid.Model.QueryCellInfo -= new GridQueryCellInfoEventHandler(Model_QueryCellInfo);
			grid.AllowSelection = saveFlags;
			grid.SortBehavior = saveBehavior;
		}

		private void gridDataBoundGrid1_CellClick(object sender, Syncfusion.Windows.Forms.Grid.GridCellClickEventArgs e)
		{
			if(e.RowIndex == 0 && e.ColIndex > 0)
			{
				//multicolumn sort
				if(SortStarting != null)
					SortStarting(this, EventArgs.Empty);
				
				int field = this.grid.Binder.ColIndexToField(e.ColIndex);
				string colName = this.grid.Binder.InternalColumns[field].MappingName;
				if ((Control.ModifierKeys & Keys.Control) == Keys.None)
				{   //normal single col sort
					//sort the clicked column by setting the SortString property
					this.SortString = colName;
				}
				else
				{ //multicolumn sort
					this.Sort(colName, false);
				}

				//redraw the grid
				this.grid.Refresh();
				if(SortCompleted != null)
					SortCompleted(this, EventArgs.Empty);
			}
		}

		private void Model_QueryCellInfo(object sender, GridQueryCellInfoEventArgs e)
		{
			// Set style.Tag for the sorted column
			if (e.RowIndex == 0 && e.ColIndex > 0)
			{
//				ListWrapper lw = this.grid.DataSource as ListWrapper;
//				if(lw != null)
//				{
					GridDataBoundGridModel grid = (GridDataBoundGridModel) sender;
					GridModelDataBinder binder = grid.Binder;
					int field = binder.ColIndexToField(e.ColIndex);
					if (field >= 0)
					{
						GridBoundColumn column = binder.InternalColumns[field];
					
						string sortName = column.MappingName;
						if (this.SortDirections.ContainsKey(sortName))
						{
							e.Style.Tag = this.SortDirections[sortName].Equals(SortOrder.Ascending) 
								? ListSortDirection.Ascending : ListSortDirection.Descending;
						}
					}
//				}
			}
		}
		#endregion

		#region events

		public event EventHandler SortStarting;
		public event EventHandler SortCompleted;

		#endregion

        #region class indexers

        public object this[int i, string propertyName]
        {
            get { return pdc[propertyName].GetValue(GetOriginalComponentAt(i)); }
            set { pdc[propertyName].SetValue(GetOriginalComponentAt(i), value); }
        }

        public object this[int i, int propertyIndex]
        {
            get { return pdc[propertyIndex].GetValue(GetOriginalComponentAt(i)); }
            set { pdc[propertyIndex].SetValue(GetOriginalComponentAt(i), value); }
        }
        #endregion

        #region List Population

        private void SetList(DataRowCollection list)
        {
            this.Clear();
            for (int i = 0; i < list.Count; ++i)
            {
                this.Add(list[i]);
            }
        }

        private void SetList(IList list)
        {
            this.Clear();
            for (int i = 0; i < list.Count; ++i)
            {
                this.Add(list[i]);
            }
        }

        public object GetOriginalComponentAt(int i)
        {
            return originalList == null ? null : originalList[i];
        }
        #endregion

        #region sorting support

		/// <summary>
		/// Sorts a column based on the column indicated in the newSortString. 
		/// </summary>
		/// <param name="newSortString">A string containing the column to be sorted. It may end with possible
		/// Asc or Desc appended to each column entry.</param>
		public void Sort(string newSortString)
        {
            Sort(newSortString, true);
        }

		/// <summary>
		/// Sorts a column based on the column indicated in the newSortString. Override this 
		/// method to manage your own sorting. For example, you could delegate
		/// the sorting to DataView.Sort if your underlying list is a dataview.
		/// </summary>
		/// <param name="newSortString">A string containing the column to be sorted. It may end with possible
		/// Asc or Desc appended to each column entry.
		/// </param>
		/// <param name="clearSort">True if you want any current sort removed before applying this sort.</param>
		/// <remarks>Setting newSortString to "Col1 ASC" would cause the current IList to be sorted 
		/// in ascending order on column 1.
		/// </remarks>
        public virtual void Sort(string newSortString, bool clearSort)
        {
//////			DataView dv = this.originalList as DataView;
//////			dv.Sort = newSortString;
//////			SetList(dv);
            //find the old sortorder if present
            string col = newSortString.ToLower().Replace(" asc", "").Replace(" desc", "");
            SortOrder lsd = sortDirections.ContainsKey(newSortString.Substring(0, col.Length))
                ? (SortOrder) sortDirections[newSortString.Substring(0, col.Length)] : SortOrder.Descending;
            //flip it
            lsd = lsd == SortOrder.Ascending ? SortOrder.Descending : SortOrder.Ascending;


            if (clearSort)
            {
                this.sortString = newSortString;//use field to avoid setter
            }
            else
            {
                //if not clearing, then check if new click is currently sorted
                //and if it is, remove it from the sortString.
                string s = newSortString + ' ' + (lsd == SortOrder.Descending ? " ASC" : " DESC"); //different order from above
                if (sortString.StartsWith(s))
                {
                    sortString = sortString.Substring(s.Length + 1);
                }
                else
                {
                    sortString = sortString.Replace(',' + s, "");
                }

                //now add the new click to the existing string
                if (sortString.Length > 0)
                    sortString += ',';
                sortString += s;
            }
            //set up the sort information arrays SortItems and SortDirections
            InitSortArrays();
            //set the proper sortorder saved above
            sortDirections[newSortString.Substring(0, col.Length)] = lsd;
            //set the sortString to reflect the sort information arrays
            SetSortString();

            //then sort things
            WrapperListComparer comparer = new WrapperListComparer(this);
            base.Sort(comparer);
        }

		

        string sortString;
        public string SortString
        {
            get { return sortString; }
            set 
            {
                if (sortString != value)
                {
                    Sort(value);
                }
            }
        }

        private string[] sortItems;
        public string[] SortItems
        {
            get { return sortItems; }
            set { sortItems = value; }
        }

        private Hashtable sortDirections = new Hashtable();
        public Hashtable SortDirections
        {
            get { return sortDirections; }
            set { sortDirections = value; }
        }

        private void SetSortString()
        {
            string s = "";
            foreach (string col in sortItems)
            {
                if (s.Length > 0)
                    s += ',';
                s += col + ' ' + (sortDirections[col].Equals(SortOrder.Ascending) ? " ASC" : " DESC");
            }

            sortString = s;
        }

        private void InitSortArrays()
        {
            string[] sorts = sortString.Split(new char[] { ',' });
            sortItems = new string[sorts.GetLength(0)];
            sortDirections.Clear();
            int i = 0;
            foreach (string s in sorts)
            {
                int k = s.LastIndexOf(' ');
                if (k == -1)
                {
                    sortItems[i] = s;
                    if (sortDirections.ContainsKey(s))
                        sortDirections[s] = SortOrder.Ascending;
                    else
                        sortDirections.Add(s, SortOrder.Ascending);
                }
                else
                {
                    sortItems[i] = s.Substring(0, k).Trim();

                    if (sortDirections.ContainsKey(s))
                        sortDirections[s] = s[k + 1].ToString().ToLower() == "a" ? SortOrder.Ascending : SortOrder.Descending;
                    else
                        sortDirections.Add(sortItems[i], s[k + 1].ToString().ToLower() == "a" ? SortOrder.Ascending : SortOrder.Descending);
                }
                i++;
            }
        }

        public void ResetList()
        {
            sortString = ""; //use field to avoid setter
            SetList(originalList);
        }

        //public int IndexInOriginalList(object o)
        //{
        //    return originalList.IndexOf(o);
        //}

        #endregion

        #region ITypedList Members

        public PropertyDescriptorCollection GetItemProperties(PropertyDescriptor[] listAccessors)
        {
            return pdc;
        }

        public string GetListName(PropertyDescriptor[] listAccessors)
        {
            return "name";
        }

        #endregion
    }

    #region IComparer class that works withs ListWrapper's sort information
    public class WrapperListComparer : IComparer
    {
         ListWrapper list;

        public WrapperListComparer(ListWrapper list)
        {
            this.list = list;
        }
        
        #region IComparer Members

        public int  Compare(object x, object y)
        {
 	        if ((x == null && y == null) || ( x == DBNull.Value && y == DBNull.Value))
                return 0;
            if (x == null || x == DBNull.Value)
                return list.SortDirections[list.SortItems[0]].Equals(SortOrder.Descending) ? -1 : 1;
            if (y == null || y == DBNull.Value)
                return list.SortDirections[list.SortItems[0]].Equals(SortOrder.Descending) ? 1 : -1;
            int c = 0;
            for (int i = 0; i < list.SortItems.GetLength(0); ++i)
            {
                IComparable ox = list.pdc[list.SortItems[i]].GetValue(x) as IComparable;
                c = ox.CompareTo(list.pdc[list.SortItems[i]].GetValue(y));
                if (list.SortDirections[list.SortItems[i]].Equals(SortOrder.Descending))
                    c = -c;
                if(c != 0)
                    break;
            }

            if(c == 0 && x != null && x != DBNull.Value)
            {
                int kk = x.GetHashCode();// list.IndexInOriginalList(x);
                int kkk = y.GetHashCode();// list.IndexInOriginalList(y);
              
                c = kk > kkk ? 1 : ((kk == kkk) ? 0 : -1);
                if (list.SortDirections[list.SortItems[list.SortItems.GetLength(0) - 1]].Equals(SortOrder.Descending))
                    c = -c;
            }
            return c;
        }

        #endregion
    }
    #endregion

}
